import { Hex } from 'viem';
import { IntentType, EcoChainIds } from '@eco-foundation/routes-ts';

declare const INTENT_EXECUTION_TYPES: readonly ["GASLESS", "SELF_PUBLISH"];
type IntentExecutionType = typeof INTENT_EXECUTION_TYPES[number];
declare const chainIds: readonly [1, 10, 130, 137, 8453, 42161, 42220, 57073];
type RoutesSupportedChainId = typeof chainIds[number];
declare const stables: readonly ["USDC", "USDbC", "USDCe", "USDT"];
type RoutesSupportedStable = typeof stables[number];
declare const stableAddresses: Record<RoutesSupportedChainId, Partial<Record<RoutesSupportedStable, Hex | undefined>>>;

type CreateSimpleIntentParams = {
    creator: Hex;
    originChainID: RoutesSupportedChainId;
    destinationChainID: RoutesSupportedChainId;
    receivingToken: Hex;
    spendingToken: Hex;
    spendingTokenLimit: bigint;
    amount: bigint;
    prover?: "HyperProver" | "StorageProver";
    recipient?: Hex;
    expiryTime?: Date;
};
type CreateIntentParams = {
    creator: Hex;
    originChainID: RoutesSupportedChainId;
    destinationChainID: RoutesSupportedChainId;
    calls: IntentCall[];
    callTokens: IntentToken[];
    tokens: IntentToken[];
    prover: "HyperProver" | "StorageProver" | Hex;
    expiryTime?: Date;
};
type IntentCall = {
    target: Hex;
    data: Hex;
    value: bigint;
};
type IntentToken = {
    token: Hex;
    amount: bigint;
};

declare class RoutesService {
    private isPreprod;
    constructor({ isPreprod }?: {
        isPreprod?: boolean;
    });
    /**
     * Creates a simple intent.
     *
     * @param {CreateSimpleIntentParams} params - The parameters for creating the simple intent.
     * @param {Hex} params.creator - The address of the intent creator.
     * @param {RoutesSupportedChainId} params.originChainID - The chain ID where the intent originates.
     * @param {RoutesSupportedChainId} params.destinationChainID - The chain ID where the intent is fulfilled.
     * @param {Hex} params.receivingToken - The token address to be received on the destination chain.
     * @param {Hex} params.spendingToken - The token address to be spent on the origin chain.
     * @param {bigint} params.spendingTokenLimit - The maximum amount of spending token to use.
     * @param {bigint} params.amount - The amount of receiving token to transfer.
     * @param {string} [params.prover="HyperProver"] - The type of prover to use.
     * @param {Hex} [params.recipient] - The recipient address (defaults to creator).
     * @param {Date} [params.expiryTime] - The expiry time for the intent (defaults to 90 minutes from now).
     * @returns {IntentType} The created intent.
     * @throws {Error} If the creator address is invalid, the origin and destination chain are the same, the amount is invalid, or the expiry time is in the past. Or if there is no prover for the specified configuration.
     */
    createSimpleIntent({ creator, originChainID, destinationChainID, receivingToken, spendingToken, spendingTokenLimit, amount, recipient, prover, expiryTime }: CreateSimpleIntentParams): IntentType;
    /**
     * Creates an intent.
     *
     * @param {CreateIntentParams} params - The parameters for creating the intent.
     * @param {Hex} params.creator - The address of the intent creator.
     * @param {RoutesSupportedChainId} params.originChainID - The chain ID where the intent originates.
     * @param {RoutesSupportedChainId} params.destinationChainID - The chain ID where the intent is fulfilled.
     * @param {IntentCall[]} params.calls - The calls to be executed on the destination chain.
     * @param {IntentToken[]} params.callTokens - The tokens required for the calls on the destination chain.
     * @param {IntentToken[]} params.tokens - The tokens to be used for rewards on the origin chain.
     * @param {string|Hex} [params.prover="HyperProver"] - The type of prover or custom prover address to use.
     * @param {Date} [params.expiryTime] - The expiry time for the intent (defaults to 90 minutes from now).
     * @returns {IntentType} The created intent.
     * @throws {Error} If the creator address is invalid, the origin and destination chain are the same, the calls or tokens are invalid, or the expiry time is in the past.
     */
    createIntent({ creator, originChainID, destinationChainID, calls, callTokens, tokens, prover, expiryTime }: CreateIntentParams): IntentType;
    /**
     * Returns the EcoChainId for a given chainId, appending "-pre" if the environment is pre-production.
     *
     * @param {RoutesSupportedChainId} chainId - The chain ID to be converted to an EcoChainId.
     * @returns {EcoChainIds} The EcoChainId, with "-pre" appended if the environment is pre-production.
     */
    getEcoChainId(chainId: RoutesSupportedChainId): EcoChainIds;
    private getProverContract;
    /**
     * Gets the address of a stable token on a specific chain.
     *
     * @param {RoutesSupportedChainId} chainID - The chain ID where the stable token is deployed.
     * @param {RoutesSupportedStable} stable - The type of stable token.
     * @returns {Hex} The hexadecimal address of the stable token.
     * @throws {Error} If the stable token is not found on the specified chain.
     */
    static getStableAddress(chainID: RoutesSupportedChainId, stable: RoutesSupportedStable): Hex;
    /**
     * Gets the stable token type from its address on a specific chain.
     *
     * @param {RoutesSupportedChainId} chainID - The chain ID where the stable token is deployed.
     * @param {Hex} address - The hexadecimal address of the stable token.
     * @returns {RoutesSupportedStable | undefined} The type of stable token.
     * @throws {Error} If no stable token is found for the given address on the specified chain.
     */
    static getStableFromAddress(chainID: RoutesSupportedChainId, address: Hex): RoutesSupportedStable | undefined;
}

type RequestQuotesForIntentParams = {
    intent: IntentType;
    intentExecutionTypes?: IntentExecutionType[];
};
type InitiateGaslessIntentParams = {
    funder: Hex;
    intent: IntentType;
    vaultAddress: Hex;
    quoteID: string;
    solverID: string;
    permitData?: PermitData;
};
type SolverQuote = {
    quoteID: string;
    solverID: string;
    quoteData: {
        quoteEntries: QuoteData[];
    };
};
type QuoteData = {
    intentExecutionType: IntentExecutionType;
    intentData: IntentType;
    expiryTime: bigint;
};
type PermitData = Permit1 | Permit2;
type Permit1 = {
    permit: {
        token: Hex;
        data: {
            signature: Hex;
            deadline: bigint;
        };
    }[];
};
type Permit2 = {
    permit2: {
        permitContract: Hex;
        permitData: SinglePermit2Data | BatchPermit2Data;
        signature: Hex;
    };
};
type SinglePermit2Data = {
    singlePermitData: {
        typedData: {
            details: Permit2DataDetails;
            spender: Hex;
            sigDeadline: bigint;
        };
    };
};
type BatchPermit2Data = {
    batchPermitData: {
        typedData: {
            details: Permit2DataDetails[];
            spender: Hex;
            sigDeadline: bigint;
        };
    };
};
type Permit2DataDetails = {
    token: Hex;
    amount: bigint;
    expiration: bigint;
    nonce: bigint;
};
type InitiateGaslessIntentResponse = {
    transactionHash: Hex;
};

declare class OpenQuotingClient {
    private readonly MAX_RETRIES;
    private dAppID;
    private axiosInstance;
    constructor({ dAppID, customBaseUrl }: {
        dAppID: string;
        customBaseUrl?: string;
    });
    /**
     * Requests quotes for a given intent.
     *
     * @param {RequestQuotesForIntentParams} params - The parameters for requesting quotes.
     * @param {IntentType} params.intent - The intent for which quotes are being requested.
     * @param {string[]} params.intentExecutionTypes - The types of intent execution for which quotes are being requested.
     * @returns {Promise<SolverQuote[]>} A promise that resolves to an array of SolverQuote objects containing the quotes.
     * @throws {Error} If intentExecutionTypes is empty or if the request fails after multiple retries.
     *
     * @remarks
     * This method sends a POST request to the `/api/v1/quotes` endpoint with the provided intent information.
     * The intentData returned in each quote will have the fee added to the reward tokens.
     */
    requestQuotesForIntent({ intent, intentExecutionTypes }: RequestQuotesForIntentParams): Promise<SolverQuote[]>;
    /**
     * Requests reverse quotes for a given intent.
     *
     * @param {RequestQuotesForIntentParams} params - The parameters for requesting reverse quotes.
     * @param {IntentType} params.intent - The intent for which quotes are being requested.
     * @param {string[]} params.intentExecutionTypes - The types of intent execution for which quotes are being requested.
     * @returns {Promise<SolverQuote[]>} A promise that resolves to an array of SolverQuote objects containing the quotes.
     * @throws {Error} If intentExecutionTypes is empty, if the calls aren't ERC20.transfer calls, or if the request fails after multiple retries.
     *
     * @remarks
     * This method sends a POST request to the `/api/v1/quotes/reverse` endpoint with the provided intent information.
     * This intentData returned in each quote will have the fee subtracted from the route tokens and calls.
     */
    requestReverseQuotesForIntent({ intent, intentExecutionTypes }: RequestQuotesForIntentParams): Promise<SolverQuote[]>;
    private formatIntentData;
    private parseQuotesResponse;
    /**
     * Initiates a gasless intent via the Open Quoting service.
     *
     * @param {InitiateGaslessIntentParams} params - The parameters for initiating a gasless intent.
     * @param {string} params.funder - The address of the entity funding the intent execution.
     * @param {IntentType} params.intent - The intent to be executed gaslessly.
     * @param {string} params.quoteID - The ID of the quote selected for execution.
     * @param {string} params.solverID - The ID of the solver that is executing the intent.
     * @param {string} params.vaultAddress - The address of the vault to use for the gasless intent.
     * @param {PermitData} [params.permitData] - Optional permit data for token approvals.
     * @returns {Promise<InitiateGaslessIntentResponse>} A promise that resolves to the response from the Open Quoting service.
     * @throws {Error} If the request fails (no retries are attempted for this endpoint).
     *
     * @remarks
     * This method sends a POST request to the `/api/v1/quotes/initiateGaslessIntent` endpoint with the provided intent information.
     */
    initiateGaslessIntent({ funder, intent, quoteID, solverID, vaultAddress, permitData }: InitiateGaslessIntentParams): Promise<InitiateGaslessIntentResponse>;
    private formatPermitData;
}

type QuoteSelectorResult = {
    quoteID: string;
    solverID: string;
    quoteData: QuoteData;
};
declare function selectCheapestQuote(solverQuotes: SolverQuote[], isReverse?: boolean, allowedIntentExecutionTypes?: IntentExecutionType[]): QuoteSelectorResult;

export { type BatchPermit2Data, type CreateIntentParams, type CreateSimpleIntentParams, INTENT_EXECUTION_TYPES, type InitiateGaslessIntentParams, type InitiateGaslessIntentResponse, type IntentExecutionType, OpenQuotingClient, type Permit1, type Permit2, type Permit2DataDetails, type PermitData, type QuoteData, type RequestQuotesForIntentParams, RoutesService, type RoutesSupportedChainId, type RoutesSupportedStable, type SinglePermit2Data, type SolverQuote, chainIds, selectCheapestQuote, stableAddresses, stables };
